/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

/**
 * Abstract Cecilia compiler Launcher.
 */
public abstract class AbstractLauncher {

  private final CmdOption          srcPathOpt   = new CmdPathOption(
                                                    new String[]{"src-path",
      "S"                                           },
                                                    "the search path of ADL,IDL and implementation files",
                                                    ".");

  private final CmdOption          outDirOpt    = new CmdOption(
                                                    new String[]{"out-path",
      "out-dir", "o"                                },
                                                    "the path where generated files will be put",
                                                    ".");

  private final CmdOption[]        coreOptions  = new CmdOption[]{srcPathOpt,
      outDirOpt                                 };

  private static final String      IDL_DIR_NAME = "idl";
  private static final String      ADL_DIR_NAME = "adl";
  private static final String      OBJ_DIR_NAME = "obj";

  private CmdOption[]              allCmbOptions;

  // Map<String, CmdOption> of options
  protected Map<String, CmdOption> options      = new HashMap<String, CmdOption>();

  // name of the ADL to compile.
  protected Map<String, String>    adlNameMap   = new TreeMap<String, String>();

  // List<String> of source files and directory
  protected List<String>           sourcePaths;
  // source class loader
  protected ClassLoader            sourceClassLoader;

  // build directory
  protected File                   buildDir;
  // idl build directory
  protected File                   idlBuildDir;
  // adl build directory
  protected File                   adlBuildDir;
  // obj build directory
  protected File                   objBuildDir;
  // external src directory
  protected File                   extSrcBuildDir;

  protected boolean                exitOnFailure;

  protected AbstractLauncher(final boolean exitOnFailure) {
    this.exitOnFailure = exitOnFailure;
  }

  protected void setSpecificOptions(final CmdOption[] specificOptions) {
    addOptions(coreOptions);
    addOptions(specificOptions);
    allCmbOptions = new CmdOption[coreOptions.length + specificOptions.length];
    System.arraycopy(coreOptions, 0, allCmbOptions, 0, coreOptions.length);
    System.arraycopy(specificOptions, 0, allCmbOptions, coreOptions.length,
        specificOptions.length);
  }

  protected abstract void printUsage(PrintStream ps);

  protected void error(final String message, final int exitVal) {
    if (exitOnFailure) {
      System.err.println(message);
      printFullUsage(System.err);
      System.exit(exitVal);
    } else {
      throw new Error(message);
    }
  }

  private void parseError(final String arg) {
    error("Unknown option '" + arg + "'", 1);
  }

  protected void internalError(final String message, final Exception e,
      final int exitVal) {
    if (exitOnFailure) {
      if (message != null)
        System.err.println("Internal error: " + message);
      else
        System.err.println("Internal error");

      if (e != null) {
        System.err.println("Caused By:");
        e.printStackTrace();
      }
      System.exit(exitVal);
    } else {
      if (e != null) {
        e.printStackTrace();
      }
      throw new InternalError(message);
    }
  }

  protected List<String> parsePathList(String paths) {
    final List<String> l = new ArrayList<String>();
    int index = paths.indexOf(File.pathSeparatorChar);
    while (index != -1) {
      l.add(paths.substring(0, index));
      paths = paths.substring(index + 1);
      index = paths.indexOf(File.pathSeparatorChar);
    }
    l.add(paths);
    return l;
  }

  protected void parseArgs(final String[] args) {
    final CmdOption execNameOption = options.get("exec-name");
    for (final String arg : args) {
      if (arg.startsWith("-")) {
        final int index = arg.indexOf('=');
        if (index < 2) {
          parseError(arg);
        }
        final String optionName = arg.substring(1, index);
        final String optionValue = arg.substring(index + 1);

        // validate option
        final CmdOption cmdOption = options.get(optionName);
        if (cmdOption == null) parseError(arg);
        cmdOption.setValue(optionValue);
      } else {
        final int index = arg.indexOf(':');
        if (index == -1) {
          // adlName = arg, no exec-name specified here, put the default value
          adlNameMap.put(arg, execNameOption.defaultValue);
        } else { // Both adlName and exec-name are specified
          final String adlName = arg.substring(0, index);
          final String execName = arg.substring(index + 1);
          adlNameMap.put(adlName, execName);
        }
      }
    }
    // Check if at least 1 adlName is specified
    if (adlNameMap.size() == 0) {
      error("no definition name is specified.", 1);
    }

    // Check the multiple exec-name definitions
    if (!execNameOption.getValue().equals(execNameOption.defaultValue)) {
      if (adlNameMap.size() == 1) {
        // Map the exec-name to the single adlName
        // The name specified with exec-name overloads the executable name
        // specified using ':' with the adlName.
        adlNameMap.put(adlNameMap.keySet().iterator().next(), execNameOption
            .getValue());
      } else {
        error(
            "-exec-name option can be specified only if there is only one archtiecture to be compiled.",
            1);
      }
    }

    prepareOutDir();
    prepareSourceClassLoader();
  }

  protected class CmdOption {
    protected final String[] options;
    protected final String   info;
    protected final String   defaultValue;
    protected String         value;

    protected CmdOption(final String[] options, final String info,
        final String defaultValue) {
      this.options = options;
      this.info = info;
      this.defaultValue = defaultValue;
    }

    protected CmdOption(final String[] options, final String info) {
      this(options, info, null);
    }

    protected CmdOption(final String option, final String info) {
      this(new String[]{option}, info, null);
    }

    protected final String getValue() {
      return (value == null) ? defaultValue : value;
    }

    protected void setValue(final String value) {
      this.value = value;
    }
  }

  protected class CmdAppendOption extends CmdOption {

    protected CmdAppendOption(final String option, final String info) {
      super(option, info);
    }

    protected CmdAppendOption(final String[] options, final String info,
        final String initialValue) {
      super(options, info, initialValue);
    }

    protected CmdAppendOption(final String[] options, final String info) {
      super(options, info);
    }

    /** Append new value at end. */
    @Override
    protected void setValue(final String value) {
      if (value == null) return;
      this.value = (this.value == null) ? value : this.value + " " + value;
    }
  }

  protected class CmdPathOption extends CmdOption {

    protected CmdPathOption(final String option, final String info) {
      super(option, info);
    }

    protected CmdPathOption(final String[] options, final String info,
        final String initialValue) {
      super(options, info, initialValue);
    }

    protected CmdPathOption(final String[] options, final String info) {
      super(options, info);
    }

    /** Append new value at end. */
    @Override
    protected void setValue(String value) {
      if (value == null) return;
      value = value.trim();
      this.value = (this.value == null) ? value : this.value
          + File.pathSeparator + value;
    }
  }

  // --------------------------------------------------------------------------
  // Utility methods
  // --------------------------------------------------------------------------

  private void addOptions(final CmdOption[] options) {
    for (final CmdOption option : options) {
      for (final String optName : option.options) {
        if (this.options.put(optName, option) != null) {
          throw new Error("Duplicated option '" + optName + "'");
        }
      }
    }
  }

  private void prepareSourceClassLoader() {
    final String srcPath = srcPathOpt.getValue();
    final List<String> paths = parsePathList(srcPath);
    sourcePaths = new ArrayList<String>(paths.size());

    // check source paths
    final Iterator<String> iterator = paths.iterator();
    while (iterator.hasNext()) {
      final String path = iterator.next();
      final File f = new File(path);
      if (!f.exists()) {
        iterator.remove();
        System.out.println("Warning '" + f.getAbsolutePath()
            + "' source path can't be found ");
        continue;
      }
      if (f.isDirectory()) {
        sourcePaths.add(path);
      } else {
        final File extractedDir = new File(extSrcBuildDir, f.getName());
        if (!extractedDir.exists() || !extractedDir.isDirectory()
            || f.lastModified() > extractedDir.lastModified()) {
          if (extractedDir.exists()) {
            System.out.println("Remove outdated external sources '"
                + extractedDir + "'");
            delTree(extractedDir);
          }
          extractedDir.mkdir();
          System.out.print("Extracting external sources '" + path + "'... ");
          System.out.flush();
          unzipFile(f, extractedDir);
          System.out.println("Done.");
        }
        sourcePaths.add(extractedDir.getAbsolutePath());
      }
    }

    // add "build/adl" and "build/idl" as source path
    sourcePaths.add(idlBuildDir.getAbsolutePath());
    sourcePaths.add(adlBuildDir.getAbsolutePath());

    // build URL array of source path
    final URL[] urls = new URL[sourcePaths.size()];
    for (int i = 0; i < urls.length; i++) {
      final String path = sourcePaths.get(i);
      final File f = new File(path);
      try {
        urls[i] = f.toURL();
      } catch (final MalformedURLException e) {
        internalError(null, e, 4);
      }
    }

    sourceClassLoader = new URLClassLoader(urls, getClass().getClassLoader());
  }

  private void prepareOutDir() {
    String outDirName = outDirOpt.getValue();
    outDirName = outDirName.trim();
    buildDir = new File(outDirName);
    if (!buildDir.exists())
      error("Invalid out directory '" + buildDir.getAbsolutePath()
          + "'. Directory not found", 2);
    if (!buildDir.isDirectory())
      error("Invalid out directory '" + buildDir.getAbsolutePath()
          + "'. Not a directory", 2);

    idlBuildDir = checkBuildDir(buildDir, IDL_DIR_NAME);
    adlBuildDir = checkBuildDir(buildDir, ADL_DIR_NAME);
    objBuildDir = checkBuildDir(buildDir, OBJ_DIR_NAME);
  }

  private File checkBuildDir(final File outDir, final String dirName) {
    final File d = new File(outDir, dirName);
    if (d.exists()) {
      if (!d.isDirectory())
        error("Invalid out directory '" + outDir.getAbsolutePath() + "'. '"
            + dirName + "' not a directory", 6);
    } else {
      if (!d.mkdir())
        error("Invalid out directory '" + outDir.getAbsolutePath()
            + "'. Unable to create'" + dirName + "' directory", 7);
    }
    return d;
  }

  private void printFullUsage(final PrintStream ps) {
    printUsage(ps);
    ps.println();
    ps
        .println("Available options are (from command line, options must be prefix with '-' and followed by '=value') :");
    int maxCol = 0;

    for (final CmdOption opt : allCmbOptions) {
      int col = 2 + opt.options[0].length();
      if (opt.options.length > 1) {
        col += opt.options[1].length() + 2;
        if (opt.options.length > 2)
          for (int j = 2; j < opt.options.length; j++) {
            col += opt.options[j].length() + 2;
          }
        else
          col += 2;
      }
      if (col > maxCol) maxCol = col;
    }
    for (final CmdOption opt : allCmbOptions) {
      final StringBuffer sb = new StringBuffer("  ");
      sb.append(opt.options[0]);
      if (opt.options.length > 1) {
        sb.append(" (");
        for (int j = 1; j < opt.options.length; j++) {
          sb.append(opt.options[j]);
          if (j < opt.options.length - 1) sb.append(", ");
        }
        sb.append(") ");
      }
      while (sb.length() < maxCol)
        sb.append(' ');
      sb.append(": ").append(opt.info);
      ps.println(sb);
    }
  }

  private void delTree(final File f) {
    if (!f.exists()) return;
    if (f.isDirectory()) {
      for (final String name : f.list()) {
        delTree(new File(f, name));
      }
    }
    if (!f.delete())
      error("Can't delete external source '" + f.getAbsolutePath() + "'", 5);
  }

  private void unzipFile(final File zipFile, final File toDir) {
    try {
      final ZipFile f = new ZipFile(zipFile);
      final Enumeration<? extends ZipEntry> enumeration = f.entries();
      final byte[] data = new byte[1024];
      while (enumeration.hasMoreElements()) {
        final ZipEntry entry = enumeration.nextElement();
        final String name = entry.getName();
        if (name.startsWith("META-INF")) continue;
        final File fEntry = new File(toDir, entry.getName());
        if (entry.isDirectory())
          fEntry.mkdirs();
        else {
          fEntry.createNewFile();
          final OutputStream os = new FileOutputStream(fEntry);
          final InputStream is = f.getInputStream(entry);
          int i = 0;
          while (true) {
            i = is.read(data);
            if (i == -1) break;
            os.write(data, 0, i);
          }
          os.close();
          is.close();
        }
      }
    } catch (final Exception e) {
      error("An error occurs while extracting external sources. '"
          + e.getMessage() + "'", 3);
    }
  }
}
