/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.cecilia.adl.ant;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Vector;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.PropertyHelper;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.taskdefs.Java;
import org.apache.tools.ant.types.Commandline;
import org.apache.tools.ant.types.EnumeratedAttribute;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.Reference;
import org.apache.tools.ant.util.ClasspathUtils;

/**
 * An Ant task that executes the ThinkADL compiler.
 */
public class ThinkADL extends Task {

  private ClasspathUtils.Delegate cpDelegate;
  private boolean                 classPathSetted = false;

  private String                  launcherClass   = "org.objectweb.fractal.cecilia.adl.Launcher";

  private final List<Argument>    args            = new ArrayList<Argument>();

  private ADLName                 adlNameNested;
  private String                  adlName;

  private Path                    srcPath;

  private File                    destDir;

  private String                  compilerADL;

  private VerboseValues           verbose;

  private Java                    j;

  @Override
  public void init() {
    cpDelegate = ClasspathUtils.getDelegate(this);
    super.init();
  }

  @Override
  public void execute() {
    // checks that required fields are setted
    if ((adlNameNested == null || adlNameNested.adlName == null)
        && adlName == null)
      throw new BuildException("You must specify an ADL to compile.");

    if (srcPath == null)
      throw new BuildException("You must specify a sourcepath.");

    if (destDir == null)
      throw new BuildException("You must specify a destdir.");

    if (!classPathSetted) {
      final Path defaultClassPath = (Path) getProject().getReference(
          "thinkadl.classpath");
      if (defaultClassPath != null) {
        cpDelegate.setClasspath(defaultClassPath);
      } else {
        throw new BuildException(
            "You must specify a classpath for the thinkADL compiler");
      }
    }

    j = new Java();
    j.setProject(getProject());
    j.setLocation(getLocation());
    j.setTaskName("thinkADL");
    Commandline.Argument jarg;
    j.setClasspath(cpDelegate.getClasspath());
    j.setClassname(launcherClass);

    j.setFork(true);
    j.setFailonerror(true);

    final List<Argument> processedArguments = new ArrayList<Argument>();
    for (final Argument arg : args) {
      processArgument(processedArguments, arg);
    }

    // set generic arguments
    for (final Argument arg : processedArguments) {
      jarg = j.createArg();
      if (arg.line != null)
        jarg.setLine(arg.line);
      else
        jarg.setValue(arg.value);
    }

    // set src path
    for (final String src : srcPath.list())
      addJarg("-src-path=" + src);

    addJarg("-out-path=" + destDir);

    if (compilerADL != null) addJarg("-compiler-adl=" + compilerADL);

    if (verbose != null)
      addJvmarg("-Dverbose.level=" + verbose.getValue().toUpperCase());

    if (adlNameNested != null)
      addJarg(adlNameNested.adlName);
    else
      addJarg(adlName);

    j.execute();
  }

  private void addJarg(final String value) {
    final Commandline.Argument jarg = j.createArg();
    jarg.setValue(value);
  }

  private void addJvmarg(final String value) {
    final Commandline.Argument jarg = j.createJvmarg();
    jarg.setValue(value);
  }

  private void processArgument(final List<Argument> processedArgList,
      final Argument argument) {
    if (argument.file != null) {
      processPropertyFile(processedArgList, argument.file);
    } else if (argument.path != null) {
      for (final String pathElem : argument.path.list()) {
        processPropertyFile(processedArgList, new File(pathElem));
      }
    } else if (argument.pathRef != null) {
      for (final String pathElem : ((Path) argument.pathRef
          .getReferencedObject()).list()) {
        processPropertyFile(processedArgList, new File(pathElem));
      }
    } else {
      processedArgList.add(argument);
    }
  }

  private void processPropertyFile(final List<Argument> processedArgList,
      final File f) {
    final Properties p = new Properties();
    try {
      p.load(new FileInputStream(f));
    } catch (final IOException e) {
      throw new BuildException("Unable to read property file '" + f.getPath()
          + "'", e);
    }
    for (final Map.Entry<Object, Object> e : p.entrySet()) {
      processProperty(processedArgList, (String) e.getKey(), (String) e
          .getValue());
    }
  }

  private void processProperty(final List<Argument> processedArgList,
      final String name, final String value) {
    final Argument processedArg;
    if (value.indexOf('$') == -1) {
      // the 'value' does not contain property references
      processedArg = new Argument(name, value);
    } else {
      // value contains property references.
      final Vector<?> fragments = new Vector<Object>();
      final Vector<?> propertyRefs = new Vector<Object>();
      PropertyHelper.getPropertyHelper(getProject()).parsePropertyString(value,
          fragments, propertyRefs);
      final StringBuilder sb = new StringBuilder();
      final Iterator<?> fragmentIter = fragments.iterator();
      final Iterator<?> propRefIter = propertyRefs.iterator();

      while (fragmentIter.hasNext()) {
        String fragment = (String) fragmentIter.next();
        if (fragment == null) {
          // fragment is a property reference.
          final String propRef = (String) propRefIter.next();

          // try to get value in project properties
          fragment = getProject().getProperty(propRef);

          if (fragment != null) sb.append(fragment);
        } else {
          sb.append(fragment);
        }
      }
      processedArg = new Argument(name, sb.toString());
    }
    processedArgList.add(processedArg);
  }

  // --------------------------------------------------------------------------
  // attribute setter methods
  // --------------------------------------------------------------------------

  /**
   * Sets the name of the ADL to compile.
   * 
   * @param adl the name of the ADL to compile.
   */
  public void setAdl(final String adl) {
    if (adlNameNested != null || adlName != null)
      throw new BuildException("ADL to compile cannot be specified twice.");
    adlName = adl;
  }

  /**
   * Sets the source path.
   * 
   * @param p the source path.
   */
  public void setSourcepath(final String p) {
    if (srcPath != null)
      throw new BuildException("Src path cannot be specified twice.");
    srcPath = new Path(getProject(), p);
  }

  /**
   * Sets the source path as a reference to a {@link Path} structure.
   * 
   * @param r the reference to the {@link Path} structure to use as source path.
   */
  public void setSourcepathref(final Reference r) {
    if (srcPath != null)
      throw new BuildException("Src path cannot be specified twice.");
    srcPath = new Path(getProject());
    srcPath.setRefid(r);
  }

  /**
   * Sets the output directory.
   * 
   * @param destDir the output directory
   */
  public void setDestdir(final File destDir) {
    if (!destDir.isDirectory())
      throw new BuildException("Invalid destdir: not a directory.");
    this.destDir = destDir;
  }

  /**
   * Sets the verbose mode of the compiler.
   * 
   * @param v the verbose mode of the compiler.
   */
  public void setVerbose(final VerboseValues v) {
    verbose = v;
  }

  /**
   * Sets the compiler's classpath as a reference to the {@link Path} structure.
   * 
   * @param r the reference to the {@link Path} structure to use as compiler's
   *            classpath.
   */
  public void setClasspathref(final Reference r) {
    cpDelegate.setClasspathref(r);
    classPathSetted = true;
  }

  /**
   * Sets the ADL of the compiler.
   * 
   * @param s the ADL of the compiler.
   */
  public void setCompileradl(final String s) {
    compilerADL = s;
  }

  /**
   * Sets the name of the class to use to launch the compiler.
   * 
   * @param c the name of the class to use to launch the compiler.
   */
  public void setLauncherclass(final String c) {
    launcherClass = c;
  }

  // --------------------------------------------------------------------------
  // Nested element factory methods
  // --------------------------------------------------------------------------

  /**
   * Creates a {@link Path} structure for the <code>classpath</code> nested
   * element. This nested element can be used to define the compiler's
   * classpath.
   * 
   * @return a {@link Path} structure for the <code>classpath</code> nested
   *         element.
   */
  public Path createClasspath() {
    classPathSetted = true;
    return cpDelegate.createClasspath();
  }

  /**
   * Creates a {@link Path} structure for the <code>src</code> nested element.
   * This nested element can be used to define the source path.
   * 
   * @return a {@link Path} structure for the <code>src</code> nested element.
   */
  public Path createSrc() {
    if (srcPath != null)
      throw new BuildException("Src path cannot be specified twice.");
    srcPath = new Path(getProject());
    return srcPath;
  }

  /**
   * Creates an {@link Argument} for <code>arg</code> nested elements. These
   * nested elements can be used to specify additional compiler arguments.
   * 
   * @return an {@link Argument} for <code>arg</code> nested elements.
   */
  public Argument createArg() {
    final Argument arg = new Argument();
    args.add(arg);
    return arg;
  }

  /**
   * Creates an {@link ADLName} for the <code>adl</code> nested element. This
   * nested element can be used to specify the name of the ADL to compile.
   * 
   * @return an {@link ADLName} for the <code>adl</code> nested element.
   */
  public ADLName createAdl() {
    if (adlNameNested != null || adlName != null)
      throw new BuildException("ADL to compile cannot be specified twice.");
    adlNameNested = new ADLName();
    return adlNameNested;
  }

  // --------------------------------------------------------------------------
  // Nested element classes
  // --------------------------------------------------------------------------

  /**
   * Class for <code>adl</code> nested element.
   */
  public static class ADLName {
    String adlName;

    /**
     * Sets the name of the ADL to compile.
     * 
     * @param adl the name of the ADL to compile.
     */
    public void setName(final String adl) {
      adlName = adl;
    }
  }

  /**
   * Class for <code>arg</code> nested element.
   * 
   * @see Commandline.Argument
   */
  public static class Argument {
    String    line;
    String    value;
    File      file;
    Path      path;
    Reference pathRef;

    protected Argument() {
    }

    protected Argument(final String name, final String value) {
      this.value = '-' + name + '=' + value;
    }

    protected boolean isProperyArgument() {
      return file != null || path != null || pathRef != null;
    }

    protected String getPropertyFile() {
      if (file != null) return file.getAbsolutePath();
      if (path != null) return path.toString();
      return pathRef.getReferencedObject().toString();
    }

    protected String[] getArgs() {
      if (value != null) return new String[]{value};
      return line.split("\\s");
    }

    /**
     * Sets the argument line.
     * 
     * @param line the argument line.
     */
    public void setLine(final String line) {
      if (file != null || value != null || path != null || pathRef != null)
        throw new BuildException(
            "you must specify only one of 'line', 'value', 'file', 'path' or 'pathref'");
      this.line = line;
    }

    /**
     * Sets the argument value.
     * 
     * @param value the argument value.
     */
    public void setValue(final String value) {
      if (line != null || file != null || path != null || pathRef != null)
        throw new BuildException(
            "you must specify only one of 'line', 'value', 'file', 'path' or 'pathref'");
      this.value = value;
    }

    /**
     * Sets the argument file. This file must be a property file.
     * 
     * @param file the argument file.
     */
    public void setFile(final File file) {
      if (line != null || value != null || path != null || pathRef != null)
        throw new BuildException(
            "you must specify only one of 'line', 'value', 'file', 'path' or 'pathref'");
      if (!file.exists())
        throw new BuildException("Invalid file name: file does not exist.");
      this.file = file;
    }

    /**
     * Sets the path to argument files. Files pointed by the given path must be
     * property files.
     * 
     * @param path the path to argument files.
     */
    public void setPath(final Path path) {
      if (line != null || value != null || file != null || pathRef != null)
        throw new BuildException(
            "you must specify only one of 'line', 'value', 'file', 'path' or 'pathref'");
      this.path = path;
    }

    /**
     * Sets the reference to path to argument files. Files pointed by the path
     * referenced by the given reference must be property files.
     * 
     * @param pathRef the reference to path to argument files.
     */
    public void setPathref(final Reference pathRef) {
      if (line != null || value != null || path != null || file != null)
        throw new BuildException(
            "you must specify only one of 'line', 'value', 'file', 'path' or 'pathref'");
      this.pathRef = pathRef;
    }
  }

  // --------------------------------------------------------------------------
  // Enumerated attribute classes
  // --------------------------------------------------------------------------

  /**
   * The class defining the possible values of the <code>verbose</code>
   * attribute.
   */
  public static class VerboseValues extends EnumeratedAttribute {

    @Override
    public String[] getValues() {
      return new String[]{"debug", "info", "error"};
    }
  }
}
