/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author:Ali Erdem Ozcan
 */

package org.objectweb.fractal.cecilia.adl.idl;

import java.io.File;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.timestamp.Timestamp;
import org.objectweb.fractal.adl.util.FractalADLLogManager;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.cecilia.adl.SourceCodeProvider;
import org.objectweb.fractal.cecilia.adl.file.SourceFileProvider;
import org.objectweb.fractal.cecilia.adl.file.SourceFileWriter;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinition;
import org.objectweb.fractal.cecilia.adl.idl.ast.InterfaceDefinition;
import org.objectweb.fractal.task.codegen.util.CodeWriter;
import org.objectweb.fractal.task.codegen.util.SourceFile;
import org.objectweb.fractal.task.core.AbstractTaskFactoryUser;
import org.objectweb.fractal.task.core.Executable;
import org.objectweb.fractal.task.core.TaskException;
import org.objectweb.fractal.task.core.primitive.annotations.ClientInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterface;
import org.objectweb.fractal.task.core.primitive.annotations.ServerInterfaces;
import org.objectweb.fractal.task.core.primitive.annotations.TaskParameters;

/**
 * Visitor component for creating virtual function definition files from
 * {@link InterfaceDefinition} nodes.
 */
public class CVFTTemplateFileVisitor extends AbstractTaskFactoryUser
    implements
      IDLDefinitionVisitor {

  // The logger
  protected static Logger logger = FractalADLLogManager.getLogger("dep");

  // ---------------------------------------------------------------------------
  // Implementation of the IDLVisitor interface
  // ---------------------------------------------------------------------------

  /**
   * Visits {@link InterfaceDefiniton} nodes, and creates a task that creates a
   * source file containing the corresponding C code.
   */
  public Component visit(final List<Node> path,
      final IDLDefinition idlDefinition, final Map<Object, Object> context)
      throws ADLException, TaskException {
    if (idlDefinition instanceof InterfaceDefinition) {
      return taskFactoryItf.newPrimitiveTask(new CVFTHeaderFileTask(
          (InterfaceDefinition) idlDefinition, (File) context
              .get("idlBuildDirectory")), (InterfaceDefinition) idlDefinition);
    } else
      return null;
  }

  // ---------------------------------------------------------------------------
  // Task classes
  // ---------------------------------------------------------------------------

  /**
   * Builds a virtual function definition file and puts the source code that is
   * received for a given interfaceNode. Provides a server interface giving
   * access to this file. Provides another server interface that returns the
   * #include file to be included in the files referencing the interface defined
   * by the given interfaceNode.
   */
  @TaskParameters("interfaceDefinitionNode")
  @ServerInterfaces({
      @ServerInterface(name = "c-vft-definition-provider", signature = SourceCodeProvider.class, record = "role:cVFTHeaderSource, id:%", parameters = {"interfaceDefinitionNode"}),
      @ServerInterface(name = "c-vft-file-provider", signature = SourceFileProvider.class, record = "role:cVFTHEaderFile, id:%", parameters = {"interfaceDefinitionNode"})})
  public static class CVFTHeaderFileTask
      implements
        Executable,
        SourceCodeProvider,
        SourceFileProvider {

    // constructor parameters
    protected final InterfaceDefinition itfDefinition;
    protected final File                idlBuildDirectory;

    // The produced source file
    protected SourceFile                sourceFile;
    // The produced source code
    protected String                    sourceCode;
    // The name of the produced source file
    protected String                    headerFileName;

    // -------------------------------------------------------------------------
    // Task client interfaces
    // -------------------------------------------------------------------------

    /** Client interface used to retrieve the content of the generated files. */
    @ClientInterface(name = "c-vft-definition", record = "role:cVFTDefinition, id:%", parameters = "interfaceDefinitionNode")
    public SourceCodeProvider           cVFTSourceItf;

    // -------------------------------------------------------------------------
    // Task constructor
    // -------------------------------------------------------------------------

    /**
     * @param itfDefinition The interface definition node for which this task
     *            must generate C source code.
     * @param idlBuildDirectory The directory into which the generated file will
     *            be placed.
     */
    public CVFTHeaderFileTask(final InterfaceDefinition itfDefinition,
        final File idlBuildDirectory) {
      this.itfDefinition = itfDefinition;
      this.idlBuildDirectory = idlBuildDirectory;
    }

    // -------------------------------------------------------------------------
    // Implementation of the Executable interface
    // -------------------------------------------------------------------------

    public void execute() throws Exception {
      headerFileName = itfDefinition.getName().replace('.', File.separatorChar)
          + "VFTTemplate.h";
      prepareSourceFile();
      prepareSourceCode();
    }

    // -------------------------------------------------------------------------
    // Implementation of the SourceCodeProvider interface
    // -------------------------------------------------------------------------

    public String getSourceCode() {
      return this.sourceCode;
    }

    // -------------------------------------------------------------------------
    // Implementation of the SourceFileProvider interface
    // -------------------------------------------------------------------------

    public SourceFile getSourceFile() {
      return this.sourceFile;
    }

    // -------------------------------------------------------------------------
    // Utility methods
    // -------------------------------------------------------------------------

    protected void prepareSourceCode() {
      final CodeWriter cw = new CodeWriter();
      cw.append("#include \"").append(headerFileName).append("\"").endl();
      sourceCode = cw.toString();
    }

    protected void prepareSourceFile() throws Exception {
      final File file = new File(idlBuildDirectory, headerFileName);
      sourceFile = new SourceFile(itfDefinition.getName(), file);

      if (file.lastModified() > Timestamp.getTimestamp(itfDefinition)) {
        if (logger.isLoggable(Level.FINE))
          logger.fine("IDL file '" + file + "' is uptodate");
      } else {
        if (!file.getParentFile().exists()) file.getParentFile().mkdirs();
        SourceFileWriter.writeToFile(file, cVFTSourceItf.getSourceCode());
      }
    }
  }
}