/***
 * Cecilia ADL Compiler
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 * Contributor: Alessio Pace
 */

package org.objectweb.fractal.cecilia.adl.interfaces;

import java.util.ArrayList;
import java.util.List;

import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;
import org.objectweb.fractal.cecilia.adl.idl.ast.IDLDefinition;

/**
 * Utility class to manipulate decorations of {@link Interface} node.
 */
public final class InterfaceDecorationUtil {

  /**
   * A decoration set on client {@link Interface} nodes to say that binding to
   * this client interface must not be established statically. The value of this
   * decoration must be a {@link Boolean}.
   * 
   * @see #setNoStaticBinding(Interface, boolean)
   * @see #isNoStaticBinding(Interface)
   * @see org.objectweb.fractal.cecilia.adl.bindings.BindingDecorationUtil#STATICALLY_BOUND_DECORATION
   */
  public static final String NO_STATIC_BINDING_DECORATION = "no-static-binding";

  /**
   * A decoration set on a client {@link Interface} node to indicates to which
   * server interface it is bound to. The value of this decoration must be an
   * {@link Interface} node.
   * 
   * @see #setBoundTo(Interface, Interface)
   * @see #getBoundTo(Interface)
   */
  public static final String BOUND_TO_DECORATION          = "bound-to";

  /**
   * A decoration set on {@link Interface} node to find the component to which
   * the interface belongs to. The value of this decoration must be an
   * {@link InterfaceContainer} node.
   */
  public static final String CONTAINER_DECORATION         = "container";

  /**
   * A decoration set on {@link IDLDefinition} node to add the ordered list of
   * IDL interfaces along the inheritance path, starting from the IDLDefinition
   * signature itself.
   */
  public static final String INHERITANCE_PATH_DECORATION  = "INHERITANCE_PATH_DECORATION";

  private InterfaceDecorationUtil() {
  }

  /**
   * Sets the {@link #NO_STATIC_BINDING_DECORATION} decoration to the given node
   * with the given boolean value.
   * 
   * @param itf the interface node to which the decoration is set.
   * @param b the value of the decoration.
   * @see #NO_STATIC_BINDING_DECORATION
   */
  public static void setNoStaticBinding(final Interface itf, final boolean b) {
    ((Node) itf).astSetDecoration(NO_STATIC_BINDING_DECORATION, b);
  }

  /**
   * Returns <code>true</code> if the given interface node has a
   * {@link #NO_STATIC_BINDING_DECORATION} decoration with the <code>true</code>
   * value.
   * 
   * @param itf an interface node.
   * @return <code>true</code> if the given binding node has a
   *         {@link #NO_STATIC_BINDING_DECORATION} decoration with the
   *         <code>true</code> value.
   * @see #NO_STATIC_BINDING_DECORATION
   */
  public static boolean isNoStaticBinding(final Interface itf) {
    final Boolean b = (Boolean) ((Node) itf)
        .astGetDecoration(NO_STATIC_BINDING_DECORATION);
    return b != null && b;
  }

  /**
   * Sets the {@link #BOUND_TO_DECORATION} decoration to the given client
   * interface node with the given server interface node value.
   * 
   * @param clientItf an interface node on which the decoration will be set.
   * @param serverItf the value of the decoration.
   * @see #BOUND_TO_DECORATION
   */
  public static void setBoundTo(final Interface clientItf,
      final Interface serverItf) {
    ((Node) clientItf).astSetDecoration(BOUND_TO_DECORATION, serverItf);
  }

  /**
   * Returns the value of the {@link #BOUND_TO_DECORATION} decoration of the
   * given client interface node.
   * 
   * @param clientItf an interface node.
   * @return The value of the {@link #BOUND_TO_DECORATION} decoration or
   *         <code>null</code> if the given node has no such decoration.
   * @see #BOUND_TO_DECORATION
   */
  public static Interface getBoundTo(final Interface clientItf) {
    return (Interface) ((Node) clientItf).astGetDecoration(BOUND_TO_DECORATION);
  }

  /**
   * Sets the value of the {@link #CONTAINER_DECORATION} decoration to the given
   * interface container node.
   * 
   * @param itf an interface node.
   * @param itfContainer the value of the decoration.
   * @see #CONTAINER_DECORATION
   */
  public static void setContainer(final Interface itf,
      final InterfaceContainer itfContainer) {
    ((Node) itf).astSetDecoration(CONTAINER_DECORATION, itfContainer);
  }

  /**
   * Returns the value of the {@link #CONTAINER_DECORATION} decoration of the
   * given interface node.
   * 
   * @param itf an interface node.
   * @return The value of the {@link #CONTAINER_DECORATION} decoration or
   *         <code>null</code> if the given node has no such decoration.
   * @see #CONTAINER_DECORATION
   */
  public static InterfaceContainer getContainer(final Interface itf) {
    return (InterfaceContainer) ((Node) itf)
        .astGetDecoration(CONTAINER_DECORATION);
  }

  public static void updateInheritancePathDecoration(final IDLDefinition node,
      final String signature) {

    List<String> inheritanceList = InterfaceDecorationUtil
        .getInheritancePathDecoration(node);

    if (inheritanceList == null) {
      inheritanceList = new ArrayList<String>();
      ((Node) node).astSetDecoration(
          InterfaceDecorationUtil.INHERITANCE_PATH_DECORATION, inheritanceList);
    }

    inheritanceList.add(signature);
  }

  public static void updateInheritancePathDecoration(final IDLDefinition node,
      final List<String> parentInheritancePath) {

    List<String> inheritanceList = InterfaceDecorationUtil
        .getInheritancePathDecoration(node);

    if (inheritanceList == null) {
      inheritanceList = new ArrayList<String>();
      ((Node) node).astSetDecoration(
          InterfaceDecorationUtil.INHERITANCE_PATH_DECORATION, inheritanceList);
    }

    inheritanceList.addAll(parentInheritancePath);
  }

  public static List<String> getInheritancePathDecoration(final Node node) {
    return (List<String>) node
        .astGetDecoration(InterfaceDecorationUtil.INHERITANCE_PATH_DECORATION);
  }

  /**
   * @param clientIDLDefinition
   * @return
   */
  public static boolean isInheritancePathDecorationPresent(
      final IDLDefinition clientIDLDefinition) {

    if (getInheritancePathDecoration(clientIDLDefinition) != null) {
      return true;
    } else {
      return false;
    }
  }
}
