/***
 * Fractal Task Codegen Framework
 * Copyright (C) 2006-2007 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.task.codegen.lib;

import java.util.Collection;
import java.util.LinkedHashSet;

import org.objectweb.fractal.task.codegen.api.FactoryProviderTask;
import org.objectweb.fractal.task.codegen.api.FileProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireFactoryProviderTask;
import org.objectweb.fractal.task.codegen.api.RequireFileProviderTask;
import org.objectweb.fractal.task.core.AbstractTask;
import org.objectweb.fractal.task.core.TaskMap;
import org.objectweb.fractal.task.core.TaskMap.TaskHole;

/**
 * Abstract implementation of {@link RequireFileProviderTask} task interface.
 * This class also implements the {@link RequireFactoryProviderTask} interface
 * so concrete task may depends on a
 * {@link FactoryProviderTask factory provider task}.
 */
public abstract class AbstractRequireFileProviderTask extends AbstractTask
    implements
      RequireFileProviderTask,
      RequireFactoryProviderTask {

  private TaskMap.TaskHole           factoryProviderTask;
  private final Collection<TaskHole> fileProviderTasks = new LinkedHashSet<TaskHole>();

  // --------------------------------------------------------------------------
  // Implementation of the RequireFactoryProviderTask interface
  // --------------------------------------------------------------------------

  public FactoryProviderTask getFactoryProviderTask() {
    return (factoryProviderTask == null)
        ? null
        : (FactoryProviderTask) factoryProviderTask.getTask();
  }

  public void setFactoryProviderTask(final TaskMap.TaskHole task) {
    if (factoryProviderTask != null) removePreviousTask(factoryProviderTask);
    factoryProviderTask = task;
    if (factoryProviderTask != null) addPreviousTask(factoryProviderTask);
  }

  // --------------------------------------------------------------------------
  // Implementation of the RequireFileProviderTask interface
  // --------------------------------------------------------------------------

  public void addFileProviderTask(final TaskMap.TaskHole task) {
    if (fileProviderTasks.add(task)) addPreviousTask(task);

  }

  public FileProviderTask[] getFileProviderTasks() {
    final FileProviderTask[] tasks = new FileProviderTask[fileProviderTasks
        .size()];
    int i = 0;
    for (final TaskMap.TaskHole hole : fileProviderTasks) {
      tasks[i++] = (FileProviderTask) hole.getTask();
    }
    return tasks;
  }

  public void removeFileProviderTask(final TaskMap.TaskHole task) {
    if (fileProviderTasks.remove(task)) removePreviousTask(task);
  }

  // --------------------------------------------------------------------------
  // Overridden Task methods
  // --------------------------------------------------------------------------

  @Override
  protected boolean doAddDependency(final TaskHole task, final Class<?> role,
      final Object context) {
    if (super.doAddDependency(task, role, context)) return true;

    if (role == RequireFileProviderTask.FILE_PROVIDER_TASK_ROLE)
      addFileProviderTask(task);
    else if (role == RequireFactoryProviderTask.FACTORY_PROVIDER_TASK_ROLE)
      setFactoryProviderTask(task);
    else
      return false;

    return true;
  }

  @Override
  protected boolean doRemoveDependency(final TaskHole task, final Class<?> role) {
    if (super.doRemoveDependency(task, role)) return true;

    if (role == RequireFileProviderTask.FILE_PROVIDER_TASK_ROLE)
      removeFileProviderTask(task);
    else if (role == RequireFactoryProviderTask.FACTORY_PROVIDER_TASK_ROLE)
      setFactoryProviderTask(null);
    else
      return false;

    return true;
  }
}
